unit uGame;
{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}
interface

uses SysUtils,
     ContNrs,

     logger,
     sdl,
     sdl_mixer,
     sdl_image,

     uBaseObjects,
     uSprite;

type
  ESDLException    = class(Exception);
  TGameApp         = class;

  TSmallIntArray   = array[0..63] of SmallInt;
  PSmallIntArray   = ^TSmallIntArray ;

  TWordArray       = array[0..63] of Word;
  PWordArray       = ^TWordArray;

  // TGame - Event, z. B. Maus oder Tastatur - Event
  TGameEvent       = procedure (const Value : Integer) of object;

  TMusicData       = class(TObject)
  protected
    fMusic            : PMix_Music;
    fOwner            : TGameApp;
    procedure Clear; dynamic;
  public
    constructor Create(aOwner : TGameApp);
    destructor Destroy; override;
    procedure LoadFromFile(const aFilename : String); dynamic;
    procedure Play; dynamic;
    property Music : PMix_Music read fMusic;
  end;

  TEffectData      = class(TObject)
    fEffect           : PMix_Chunk;
    fOwner            : TGameApp;
    procedure Clear; dynamic;
  public
    constructor Create(aOwner : TGameApp);
    destructor Destroy; override;
    procedure LoadFromFile(const aFilename : String); dynamic;
    procedure Play; dynamic;
    property Effect : PMix_Chunk read fEffect;
  end;

  TBaseJoystick    = class(TObject)
  protected
    fOwner           : TGameApp;
    fDevice          : PSDL_Joystick;
    fActive          : Boolean;
    fIndex           : Integer;
    function  GetName : String; virtual; abstract;
    function  GetButtonCount : Integer; virtual; abstract;
    procedure SetActive(aData : Boolean); virtual; abstract;
    function  GetAxisX : Double; virtual; abstract;
    function  GetAxisY : Double; virtual; abstract;
    function  GetAxisZ : Double; virtual; abstract;
    function  GetButtonPressed(aButton : Integer) : Boolean; virtual; abstract;
  public
    constructor Create(aOwner : TGameApp; aIndex : Integer); virtual;

    procedure Open; virtual; abstract;
    procedure Close; virtual; abstract;

    property Actice : Boolean read fActive write SetActive;
    property AxisX : Double read GetAxisX;
    property AxisY : Double read GetAxisY;
    property AxisZ : Double read GetAxisZ;
    property ButtonCount : Integer read GetButtonCount;
    property ButtonPressed[aButton : Integer] : Boolean read GetButtonPressed;
    property Index : Integer read fIndex;
    property Name : String read GetName;
    property Owner : TGameApp read fOwner;
  end;

  TJoystick        = class(TBaseJoystick)
  protected
    fOwner           : TGameApp;
    fDevice          : PSDL_Joystick;
    fActive          : Boolean;
    fIndex           : Integer;
    function  GetName : String; override;
    function  GetButtonCount : Integer; override;
    procedure SetActive(aData : Boolean); override;
    function  GetAxisX : Double; override;
    function  GetAxisY : Double; override;
    function  GetAxisZ : Double; override;
    function  GetButtonPressed(aButton : Integer) : Boolean; override;
  public
    procedure Open; override;
    procedure Close; override;

    property Actice : Boolean read fActive write SetActive;
    property AxisX : Double read GetAxisX;
    property AxisY : Double read GetAxisY;
    property AxisZ : Double read GetAxisZ;
    property ButtonCount : Integer read GetButtonCount;
    property ButtonPressed[aButton : Integer] : Boolean read GetButtonPressed;
    property Name : String read GetName;
    property Index : Integer read fIndex;
  end;

  TDummyJoystick   = class(TBaseJoystick)
  protected
    function  GetName : String; override;
    function  GetButtonCount : Integer; override;
    procedure SetActive(aData : Boolean); override;
    function  GetAxisX : Double; override;
    function  GetAxisY : Double; override;
    function  GetAxisZ : Double; override;
    function  GetButtonPressed(aButton : Integer) : Boolean; override;
  public
    procedure Open; override;
    procedure Close; override;
  end;

  TJoysticks       = class(TObject)
  protected
    fOwner           : TGameApp;
    fList            : TObjectlist;
    fUseEventNotify  : Boolean;
    function  GetCount : Integer;
    function  GetJoystick(aIndex : Integer) : TBaseJoystick;
    procedure SetUseEventNotify(aState : Boolean);
  public
    constructor Create(aOwner : TGameApp);
    destructor Destroy; override;
    procedure EnumJoysticks;
    procedure OpenAll;
    procedure CloseAll;
    property Count : Integer read GetCount;
    property Items[aIndex : Integer] : TBaseJoystick read GetJoystick; default;
    property UseEventNotify : Boolean read fUseEventNotify write SetUseEventNotify;
  end;

  TGameState       = (
                       gsMenue,
                       gsPlaying,
                       gsPaused,
                       gsGameOver
                     );
  TGameStates      = set of TGameState;

  TGameApp         = class(TObject)
  private
    fTerminated      : Boolean;
    fInitDone        : Boolean;
    fShowCursor      : Boolean;
    fGameState       : TGameState;
    fGameTime        : Integer;
    fFrameTime       : Integer;

    fKeys            : PUInt8Array;
    fModState        : TSDLMod;
    fJoysticks       : TJoysticks;
    procedure SetShowCursor(aData : Boolean);
  protected
    fCaption         : String;
    fObjects         : TObjectList;
    fSprites         : TObjectList;

    // Video - Data
    fWidth           : Integer;
    fHeight          : Integer;
    fFullScreen      : Boolean;

    // Audio - Data
    fChannels        : Integer;
    fFrequenz        : Integer;
    fAudioFormat     : UInt16;
    fChunksize       : Integer;

    fBackgroundMusic : TMusicData;

    fScreen          : PSDL_Surface;
  public
    constructor Create(const aCaption : String);
    destructor Destroy; override;

    procedure RaiseLastSDLError(const aMsg : String = '');

    procedure SetupSDL;
    procedure SetupVideo; dynamic;
    procedure SetupAudio; dynamic;

    function  LoadImage(const aFilename : String) : PSDL_Surface;
    function  LoadMusic(const aFilename : String) : TMusicData;
    function  LoadSound(const aFilename : String) : TEffectData;

    function  LoadSprite(aBase : TSpriteBase; aX, aY, aSpeed : Integer) : TSprite; overload;
    function  LoadSprite(aBase : TSpriteBase; aX, aY : Integer) : TSprite; overload;
    function  LoadSprite(aBase : TSpriteBase) : TSprite; overload;

    procedure DrawIMG(Img : PSDL_Surface; x, y : Integer); overload;
    procedure DrawIMG(Img : PSDL_Surface; x, y, w, h, x2, y2 : Integer); overload;

    procedure GameInit; dynamic;

    procedure SwitchToMenueState;
    procedure SwitchToPlaying;
    procedure SwitchToPaused;
    procedure SwitchToGameOver;
    function  IsState(aState : TGameState) : Boolean; overload;
    function  IsState(aStates : TGameStates): Boolean; overload;

    function  HandleKeyEvent(var Key : TSDL_KeyboardEvent) : Boolean; dynamic;
    function  HandleJoystickEvent(const aEvent : TSDL_Event) : Boolean; dynamic;
    function  HandleApplicationEvent(const aEvent : TSDL_Event) : Boolean; dynamic;
    function  HandleEvents(var aEvent : TSDL_Event) : Boolean; dynamic;

    procedure BeforeDrawScene; dynamic;
    procedure DrawScene; dynamic;
    procedure AfterDrawScene; dynamic;

    procedure GameShutdown; dynamic;

    procedure Run; dynamic;

    procedure Quit; dynamic;

    property BackgroundMusic : TMusicData read  fBackgroundMusic
                                          write fBackgroundMusic;
    property FrameTime : Integer read fFrameTime;
    property FullScreen : Boolean read fFullScreen write fFullScreen;
    property GameState : TGameState read fGameState;
    property GameTime : Integer read fGameTime;
    property Height : Integer read fHeight write fHeight;
    property Joysticks : TJoysticks read fJoysticks;
    property Keys : PUInt8Array read fKeys;
    property ModState : TSDLMod read fModState;

    property Widht : Integer read fWidth write fWidth;
    property Screen : PSDL_Surface read fScreen;
    property ShowCursor : Boolean read fShowCursor write SetShowCursor;
    property Terminated : Boolean read fTerminated;
  end;

implementation

// TMusicData ------------------------------------------------------------
procedure TMusicData.Clear;
begin
  if Assigned(fMusic) then begin
    Mix_FreeMusic(fMusic);
    fMusic := nil;
  end;
end;

constructor TMusicData.Create(aOwner : TGameApp);
begin
  inherited Create;

  fOwner := aOwner;
  fMusic := nil;
end;

destructor TMusicData.Destroy;
begin
  Clear;

  inherited Destroy;
end;

procedure TMusicData.LoadFromFile(const aFilename : String);
var
  Buffer           : array[0..255] of Char;
begin
  Clear;

  StrPLCopy(Buffer, aFilename, SizeOf(Buffer) - 1);
  fMusic := Mix_LoadMUS(Buffer);
  if not Assigned(fMusic) then begin
    fOwner.RaiseLastSDLError();
  end;
end;

procedure TMusicData.Play;
begin
  if (Mix_PlayingMusic() = 0) then begin
    Mix_PlayMusic(fMusic, 0);
  end;
end;

// TMusicData ------------------------------------------------------------
procedure TEffectData.Clear;
begin
  if Assigned(fEffect) then begin
    Mix_FreeChunk(fEffect);
    fEffect := nil;
  end;
end;

constructor TEffectData.Create(aOwner : TGameApp);
begin
  inherited Create;

  fOwner  := aOwner;
  fEffect := nil;
end;

destructor TEffectData.Destroy;
begin
  Clear;

  inherited Destroy;
end;

procedure TEffectData.LoadFromFile(const aFilename : String);
var
  Buffer           : array[0..255] of Char;
begin
  Clear;

  StrPLCopy(Buffer, aFilename, SizeOf(Buffer) - 1);
  fEffect := Mix_LoadWAV(Buffer);
  if not Assigned(fEffect) then begin
    fOwner.RaiseLastSDLError();
  end;
end;

procedure TEffectData.Play;
begin
  if Assigned(fEffect) then begin
    Mix_PlayChannel(1, fEffect, 0);
  end;
end;

//  TAbstractJoystick ----------------------------------------------------
constructor TBaseJoystick.Create(aOwner : TGameApp; aIndex : Integer);
begin
  inherited Create;

  fOwner  := aOwner;
  fActive := False;
  fIndex  := aIndex;
end;

//  TJoystick ------------------------------------------------------------
procedure TJoystick.SetActive(aData : Boolean);
begin
  if (fActive <> aData) then begin
    if aData then begin
      Open;
    end else begin
      Close;
    end;
  end;
end;

function TJoystick.GetName : String;
begin
  Result := SDL_JoystickName(fIndex);
end;

function TJoystick.GetButtonCount : Integer;
begin
  Result := 0;
  if Assigned(fDevice) then begin
    Result := SDL_JoystickNumButtons(fDevice);
  end;
end;

function TJoystick.GetAxisX : Double;
var
  iState           : SmallInt;
begin
  Result := 0.0;
  if Assigned(fDevice) then begin
    if (fDevice^.nAxes >= 1) then begin
      iState := PSMALLINTARRAY(fDevice^.Axes)^[0];
      Result := (iState / 32768.0);
    end;
  end;
end;

function TJoystick.GetAxisY : Double;
var
  iState           : SmallInt;
begin
  Result := 0.0;
  if Assigned(fDevice) then begin
    if (fDevice^.nAxes >= 2) then begin
      iState := PSMALLINTARRAY(fDevice^.Axes)^[1];
      Result := (iState / 32768.0);
    end;
  end;
end;

function TJoystick.GetAxisZ : Double;
var
  iState           : SmallInt;
begin
  Result := 0.0;
  if Assigned(fDevice) then begin
    if (fDevice^.nAxes >= 3) then begin
      iState := PSMALLINTARRAY(fDevice^.Axes)^[2];
      Result := (iState / 32768.0);
    end;
  end;
end;

function TJoystick.GetButtonPressed(aButton : Integer) : Boolean;
begin
  Result := False;
  if Assigned(fDevice) then begin
    if (aButton < fDevice^.nButtons) then begin
      Result := (PWORDARRAY(fDevice^.Buttons)^[aButton] = 1);
    end;
  end;
end;

procedure TJoystick.Open;
begin
  if not fActive then begin
    fDevice := SDL_JoystickOpen(fIndex);
    if not Assigned(fDevice) then begin
      Log.LogStatus(StrPas(SDL_GetError()), 'TJoystick.Open');
    end else begin
      fActive := True;
    end;
  end;
end;

procedure TJoystick.Close;
var
  Temp             : PSDL_Joystick;
begin
  if not fActive then begin
    fActive := False;
    Temp    := fDevice;
    fDevice := nil;
    SDL_JoystickClose(Temp);
  end;
end;

// TDummyJoystick --------------------------------------------------------
function TDummyJoystick.GetName : String;
begin
  Result := 'Keyboard to Joytick - Converter';
end;

function TDummyJoystick.GetButtonCount : Integer;
begin
  Result := 2;
end;

procedure TDummyJoystick.SetActive(aData : Boolean);
begin
  fActive := aData;
end;

function TDummyJoystick.GetAxisX : Double;
begin
  Result := 0.0;
  if (Owner.Keys[SDLK_LEFT] <> 0) then begin
    Result := -1.0;
  end else if (Owner.Keys[SDLK_RIGHT] <> 0) then begin
    Result :=  1.0;
  end;
end;

function TDummyJoystick.GetAxisY : Double;
begin
  Result := 0.0;
  if (Owner.Keys[SDLK_UP] <> 0) then begin
    Result := -1.0;
  end else if (Owner.Keys[SDLK_DOWN] <> 0) then begin
    Result :=  1.0;
  end;
end;

function TDummyJoystick.GetAxisZ : Double;
begin
  Result := 0.0;
end;

function TDummyJoystick.GetButtonPressed(aButton : Integer) : Boolean;
begin
  case aButton of
    0 : Result := (Owner.Keys[SDLK_SPACE] <> 0);
  else
    Result := False;
  end;
end;

procedure TDummyJoystick.Open;
begin
  fActive := True;
end;

procedure TDummyJoystick.Close;
begin
  fActive := False;
end;

// TJoysticks ------------------------------------------------------------
function TJoysticks.GetCount : Integer;
begin
  Result := fList.Count;
end;

function TJoysticks.GetJoystick(aIndex : Integer) : TBaseJoystick;
begin
  Result := (fList[aIndex] as TBaseJoystick);
end;

procedure TJoysticks.SetUseEventNotify(aState : Boolean);
begin
  if (fUseEventNotify <> aState) then begin
    if aState then begin
      SDL_JoystickEventState(SDL_ENABLE);
    end else begin
      SDL_JoystickEventState(SDL_IGNORE);
    end;
    fUseEventNotify := aState;
  end;
end;

constructor TJoysticks.Create(aOwner : TGameApp);
begin
  inherited Create;

  fOwner          := aOwner;
  fList           := TObjectlist.Create;
  fUseEventNotify := True;
end;

destructor TJoysticks.Destroy;
begin
  FreeAndNil(fList);

  inherited Destroy;
end;

procedure TJoysticks.EnumJoysticks;
var
  iLoop            : Integer;
begin
  fList.Clear;
  for iLoop := 0 to Pred(SDL_NumJoysticks) do begin
    fList.Add(TJoystick.Create(fOwner, iLoop));
  end;

  if (fList.Count = 0) then begin
    // Keinen Joystick gefunden, erstelle einen Dummy - Joystick,
    // welches die Tastatursignale in Joystick - Signale wandelt.
    fList.Add(TDummyJoystick.Create(fOwner, 0));
  end;
end;

procedure TJoysticks.OpenAll;
var
  iLoop            : Integer;
begin
  for iLoop := 0 to Pred(Count) do begin
    Self[iLoop].Close;
  end;
end;

procedure TJoysticks.CloseAll;
var
  iLoop            : Integer;
begin
  for iLoop := 0 to Pred(Count) do begin
    if Self[iLoop].Actice then begin
      Self[iLoop].Close;
    end;
  end;
end;


// TGameApp --------------------------------------------------------------
procedure TGameApp.SetShowCursor(aData : Boolean);
begin
  if (aData <> fShowCursor) then begin
    SetupSDL;
    fShowCursor := aData;
    if fShowCursor then begin
      SDL_ShowCursor(1);
    end else begin
      SDL_ShowCursor(0);
    end;
  end;
end;

procedure TGameApp.RaiseLastSDLError;
begin
  if (aMsg <> '') then begin
    raise ESDLException.CreateFmt('%s: %s\n', [aMsg, StrPas(SDL_GetError())]);
  end else begin
    raise ESDLException.Create(StrPas(SDL_GetError()));
  end;
end;

constructor TGameApp.Create(const aCaption : String);
begin
  inherited Create;

  fInitDone        := False;
  fTerminated      := True;
  fShowCursor      := True;
  fCaption         := aCaption;
  fBackgroundMusic := nil;
  fObjects         := TObjectList.Create;
  fJoysticks       := TJoysticks.Create(Self);

  // Video settings
  fWidth      := 640;
  fHeight     := 480;
  fScreen     := nil;
  fFullScreen := False;

  // Audio settings
  fFrequenz    := 44100;
  fAudioFormat := AUDIO_S16SYS;
  fChannels    := 2;
  fChunksize   := 2048;

  // App settings
  fGameTime   := 0;
end;

destructor TGameApp.Destroy;
begin
  FreeAndNil(fObjects);
  FreeAndNil(fJoysticks);
  
  if fInitDone then begin
    Quit;
  end;

  inherited Destroy;
end;

procedure TGameApp.SetupSDL;
var
  cBuffer          : array[0..255] of Char;
  iloop            : Integer;
begin
  if not fInitDone then begin
    // Initalize the SDL Video, Audio and Joystick subsystems.
    if (SDL_Init(SDL_INIT_AUDIO or SDL_INIT_VIDEO or SDL_INIT_JOYSTICK) < 0)
    then begin
      RaiseLastSDLError('Unable to init SDL');
    end;

    // Fenstertitel festlegen
    StrPLCopy(cBuffer, fCaption, Pred(SizeOf(cBuffer)));
    SDL_WM_SetCaption(cBuffer, nil);

    Log.LogStatus('Enumerate joysticks', 'uGame.SetupSDL');
    fJoysticks.EnumJoysticks;
    Log.LogStatus(Format('Found %d Joysticks', [fJoysticks.Count]), '');
    for iLoop := 0 to Pred(fJoysticks.Count) do begin
      Log.LogStatus(Format('Joystick %d: "%s"', [iLoop + 1, fJoysticks[iLoop].Name]), '');
    end;
    fInitDone  := True;
  end;
end;

procedure TGameApp.SetupVideo;
var
  iFlags           : Cardinal;
begin
  SetupSDL;

  // SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_FULLSCREEN|SDL_ANYFORMAT

  iFlags := SDL_SWSURFACE;
  if fFullScreen then begin
    iFlags := iFlags or SDL_FULLSCREEN;
  end;

  Log.LogStatus('Setup video mode', 'TGameApp.SetupVideo');

  fScreen := SDL_SetVideoMode(fWidth, fHeight, 32, iFlags);
  if not Assigned(fScreen) then begin
    RaiseLastSDLError('Unable to set %dx%d video');
  end;

  Log.LogStatus('video mode initialization done', 'TGameApp.SetupVideo');
end;

procedure TGameApp.SetupAudio;
begin
  if (Mix_OpenAudio(fFrequenz, fAudioFormat, fChannels, fChunksize) < 0)
  then begin
    RaiseLastSDLError(Format('Couldn''t set %d Hz audio\n- Reason', [fFrequenz]));
  end;
end;

function TGameApp.LoadImage(const aFilename : String) : PSDL_Surface;
var
  temp1, temp2     : PSDL_Surface;
  cFile            : array[0..255] of Char;
begin
  Log.LogStatus('Loading "' + aFilename + '"', 'uGame.ImageLoad');
  StrPLCopy(cFile, aFilename, Pred(SizeOf(cFile)));
  if (Pos('.jpg', Lowercase(aFilename)) > 0) then begin
    temp1 := IMG_LoadJPG_RW(SDL_RWFromFile(cFile, 'rb'));
  end else begin
    temp1 := SDL_LoadBMP_RW(SDL_RWFromFile(cFile, 'rb'), 1);
  end;
  temp2 := SDL_DisplayFormat(temp1);
  SDL_FreeSurface(temp1);

  Log.LogStatus('"' + aFilename + '" loaded', 'uGame.ImageLoad');
  Result := temp2;
end;

function TGameApp.LoadMusic(const aFilename : String) : TMusicData;
begin
  Result := TMusicData.Create(Self);
  try
    Log.LogStatus('Loading "' + aFilename + '"', 'uGame.LoadMusic');
    Result.LoadFromFile(aFilename);
    fObjects.Add(Result);
    Log.LogStatus('"' + aFilename + '" loaded', 'uGame.LoadMusic');
  except
    FreeAndNil(Result);
    raise;
  end;
end;

function TGameApp.LoadSound(const aFilename : String) : TEffectData;
begin
  Result := TEffectData.Create(Self);
  try
    Log.LogStatus('Loading "' + aFilename + '"', 'uGame.LoadSound');
    Result.LoadFromFile(aFilename);
    fObjects.Add(Result);
    Log.LogStatus('"' + aFilename + '" loaded', 'uGame.LoadSound');
  except
    FreeAndNil(Result);
    raise;
  end;
end;

function TGameApp.LoadSprite(aBase : TSpriteBase; aX, aY, aSpeed : Integer) : TSprite;
begin
  Result := TSprite.Create;
  if Assigned(Result) then begin
    Result.Init(aBase, Screen);
    Result.setXY(aX, aY);
    Result.AnimSpeed := aSpeed;
  end;
end;

function TGameApp.LoadSprite(aBase : TSpriteBase; aX, aY : Integer) : TSprite;
begin
  Result := LoadSprite(aBase, aX, aY, 0);
end;

function TGameApp.LoadSprite(aBase : TSpriteBase) : TSprite;
begin
  Result := LoadSprite(aBase, 0, 0, 0);
end;

// Blit an image on the screen surface
procedure TGameApp.DrawIMG(Img : PSDL_Surface; x, y : Integer);
var
  Dest             : TSDL_Rect;
begin
  Dest.x := X;
  Dest.y := Y;
  SDL_BlitSurface(Img, nil, Screen, @Dest);
end;

// Blit a portion of an image on the screen surface
procedure TGameApp.DrawIMG(Img : PSDL_Surface; x, y, w, h, x2, y2 : Integer);
var
  Dest1            : TSDL_Rect;
  Dest2            : TSDL_Rect;
begin
  Dest1.x := x;
  Dest1.y := y;
  Dest2.x := x2;
  Dest2.y := y2;
  Dest2.w := w;
  Dest2.h := h;
  SDL_BlitSurface(Img, @Dest2, Screen, @Dest1);
end;

procedure TGameApp.GameInit;
begin
end;

procedure TGameApp.SwitchToMenueState;
begin
  fGameState := gsMenue;
end;

procedure TGameApp.SwitchToPlaying;
begin
  fGameState := gsPlaying;
end;

procedure TGameApp.SwitchToPaused;
begin
  fGameState := gsPaused;
end;

procedure TGameApp.SwitchToGameOver;
begin
  fGameState := gsGameOver;
end;

function TGameApp.IsState(aState : TGameState) : Boolean;
begin
  Result := (fGameState = aState);
end;

function TGameApp.IsState(aStates : TGameStates): Boolean;
begin
  Result := (fGameState in aStates);
end;

function TGameApp.HandleKeyEvent(var Key : TSDL_KeyboardEvent) : Boolean;
begin
  Result := False;
end;

function TGameApp.HandleJoystickEvent(const aEvent : TSDL_Event) : Boolean;
begin
  Result := False;
end;

function TGameApp.HandleApplicationEvent(const aEvent : TSDL_Event) : Boolean;
begin
  Result := False;
end;

function TGameApp.HandleEvents(var aEvent : TSDL_Event) : Boolean;
begin
  Result := False;

  // Has a key been pressed down?
  if (aEvent.Type_ = SDL_KEYDOWN) then begin
    if HandleKeyEvent(aEvent.Key) then Exit;

    if (aEvent.Key.Keysym.sym = SDLK_ESCAPE) then begin
      fTerminated := True;
      Result      := True;
      Exit;
    end;
  end;
end;

procedure TGameApp.BeforeDrawScene;
begin
end;

procedure TGameApp.DrawScene;
begin
end;

procedure TGameApp.AfterDrawScene;
begin
end;

procedure TGameApp.GameShutdown;
begin
  fJoysticks.CloseAll;
end;

procedure TGameApp.Run;
var
  Event            : TSDL_Event;
  wLastTickCount   : UInt32;
  wTickCount       : UInt32;
begin
  // Eintritt in Main-Loop
  fTerminated := False;

  SwitchToMenueState;
  fGameState     := gsGameOver;
  wLastTickCount := 0;
  fGameTime      := 0;
  fFrameTime     := 0;
  while not fTerminated do begin
    if not Joysticks.UseEventNotify then begin
      // Es die SDL Event - Steuerung nicht verwendet, daher muessen
      // die Joysticks manuelle aufgerufen werden
      SDL_JoystickUpdate;
    end;

    if (SDL_PollEvent(@Event) <> 0) then begin
      HandleEvents(Event);
    end;

    // Get the state of the keyboard keys
    fKeys          := PUInt8Array(SDL_GetKeyState(nil));
    fModState      := SDL_GetModState;

    // calculate time variables
    wTickCount     := SDL_GetTicks();
    fFrameTime     := wTickCount - wLastTickCount;
    wLastTickCount := wTickCount;
    if (GameState = gsPlaying) then begin
      fGameTime := fGameTime + fFrameTime;
    end;

    BeforeDrawScene;

    DrawScene;

    if Assigned(fBackgroundMusic) then begin
      fBackgroundMusic.Play;
    end;

    AfterDrawScene;
  end;

  GameShutdown;
end;

procedure TGameApp.Quit;
begin
  SDL_QUIT;
end;

end.

