unit uSprite;
(*
  Convert from c++ to delphi and made an object.

  sprite.cpp
  Cone3D SDL font routines.
  Made by Marius Andra 2002
  http://cone3d.gamedev.net

  You can use the code for anything you like.
  Even in a commercial project.
  But please let me know where it ends up.
  I'm just curious. That's all.
*)
{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

interface

uses Types,
     SysUtils,

     Logger,
     SDL,
     SDL_Image,
     SDLUtils;

type
  TSprite          = class;

  TSpriteFrame     = class(TObject)
  protected
    fImage           : PSDL_Surface;
    fPause           : Integer;
  public
    constructor Create(aFilename : PChar; aPause, aRed, aGreen, aBlue : Integer);
    destructor Destroy; override;
    property Image : PSDL_Surface read fImage;
    property Pause : Integer read fPause;
  end;

  TAnimation       = array of TSpriteFrame;

  TSpriteBase      = class(TObject)
  private
    fAnim            : TAnimation;
    fBuilt           : Integer;
    fNumframes       : Integer;
    fW               : Integer;
    fH               : Integer;
  public
    destructor Destroy; override;
    function Init(const aDir : String) : Integer; dynamic;
    property Anim : TAnimation read fAnim;
    property Height : Integer read fH;
    property NumFrames : Integer read fNumFrames;
    property Width : Integer read fW;
  end;

  TSpriteType      = (
                       stEnemy,
                       stShoot
                     );

  TAnimState       = (
                       asNone,
                       asAnimStart,
                       asAnimNext,
                       asAnimStop
                     );

  TAnimStateNotify = procedure(aSprite : TSprite; aState : TAnimState) of Object;

  TSprite          = class(TObject)
  private
    fLocked          : Integer;
    fTag             : Integer;
    fData            : Pointer;
    fFrame           : Integer;
    fX, fY           : Double;
    fOldX, fOldY     : Double;
    fAnimating       : Boolean;
    fDrawn           : Integer;
    fAnimSpeed       : Double;
    fMoveSpeed       : Double;
    fLives           : Integer;
    fLastUpdate      : Integer;
    fSpriteBase      : TSpriteBase;
    fBackreplacement : PSDL_Surface;
    fScreen          : PSDL_Surface;
    fAnimState       : TAnimState;
    fAnimStateChange : TAnimStateNotify;
    fVisible         : Boolean;
    procedure SetAmimState(aState : TAnimState);
    function GetWidth : Integer;
    function GetHeight : Integer;
    function GetRect : Types.TRect;
    function GetSDLRect : TSDL_Rect;
    function GetCenterPos : TPoint;
  public
    constructor Create;
    destructor Destroy; override;

    function  Init(aBase : TSpriteBase; aScreen : PSDL_Surface) : Integer;
    procedure Draw;
    procedure clearBG;
    procedure updateBG;

    procedure ToggleAnim;
    procedure StartAnim;
    procedure StopAnim;
    procedure Rewind;

    function  SpriteHit(aSprite : TSprite) : Boolean;
    function  SpriteHitPixel(aSprite : TSprite) : Boolean;

    procedure SetCenter(X, Y : Double); overload;
    procedure SetCenter(Pos : TPoint); overload;

    procedure SetXY(X, Y : Double);
    procedure XAdd(dX : Double);
    procedure YAdd(dY : Double);

    property AnimSpeed : Double read fAnimSpeed write fAnimSpeed;
    property AnimState : TAnimState read fAnimState write SetAmimState;
    property CenterPos : TPoint read GetCenterPos;
    property Data : Pointer read fData write fData;
    property Frame : Integer read fFrame write fFrame;
    property Height : Integer read GetHeight;
    property MoveSpeed : Double read fMoveSpeed write fMoveSpeed;
    property OnAnimStateChange : TAnimStateNotify read  fAnimStateChange
                                                  write fAnimStateChange;
    property Rect : Types.TRect read GetRect;
    property SDLRect : TSDL_Rect read GetSDLRect;
    property SpriteBase : TSpriteBase read fSpriteBase;
    property Tag : Integer read fTag write fTag;
    property Width : Integer read GetWidth;
    property Visible : Boolean read fVisible write fVisible;
    property X : Double read fX write fX;
    property Y : Double read fY write fY;
  end;

implementation

uses usScanf;

function LoadImage(const aFilename : String) : PSDL_Surface;
var
  temp1, temp2     : PSDL_Surface;
  cFile            : array[0..255] of Char;
begin
  Log.LogStatus('Loading "' + aFilename + '"', 'uGame.ImageLoad');
  StrPLCopy(cFile, aFilename, Pred(SizeOf(cFile)));
  if (Pos('.jpg', Lowercase(aFilename)) > 0) then begin
    temp1 := IMG_LoadJPG_RW(SDL_RWFromFile(cFile, 'rb'));
  end else if (Pos('.tga', Lowercase(aFilename)) > 0) then begin
    temp1 := IMG_LoadTGA_RW(SDL_RWFromFile(cFile, 'rb'));
  end else if (Pos('.bmp', Lowercase(aFilename)) > 0) then begin
    Result := SDL_LoadBMP(cFile);
    Exit;
  end else begin
    Result := nil;
    Exit;
  end;
  temp2 := SDL_DisplayFormat(temp1);
  SDL_FreeSurface(temp1);

  Log.LogStatus('"' + aFilename + '" loaded', 'uGame.ImageLoad');
  Result := temp2;
end;

// -----------------------------------------------------------------------
constructor TSpriteFrame.Create(aFilename : PChar; aPause, aRed, aGreen, aBlue : Integer);
var
  Img              : PSDL_Surface;
begin
  inherited Create;

  Img := LoadImage(aFilename);
  try
    if (Img = nil) then begin
      StrPas(SDL_GetError());
      Exit;
    end;

    if (aRed >= 0) or (aGreen >= 0) or (aBlue >= 0) then begin
      SDL_SetColorKey(Img, SDL_SRCCOLORKEY, SDL_MapRGB(Img.Format, aRed, aGreen, aBlue));
    end;

    fImage := SDL_DisplayFormat(Img);
    fPause := aPause;
  finally
    SDL_FreeSurface(Img);
  end;
end;

destructor TSpriteFrame.Destroy;
begin
  SDL_FreeSurface(fImage);

  inherited Destroy;
end;

// -----------------------------------------------------------------------
destructor TSpriteBase.Destroy;
var
  iLoop            : Integer;
begin
  for iLoop := 0 to Pred(Length(fAnim)) do begin
    fAnim[iLoop].Free;
  end;

  SetLength(fAnim, 0);

  inherited Destroy;
end;

function TSpriteBase.Init(const aDir : String) : Integer;
var
  Buffer           : String;
  Filename         : array[0..255] of Char;
  name             : String;
  Count            : Integer;
  pause            : Integer;
  r, g, b          : Integer;
  fp               : System.Text;
begin
  Result := -1;
  pause  := 0;
  r      := 0;
  g      := 0;
  b      := 0;

  AssignFile(fp, aDir + '/info');
  Reset(fp);
  try
    ReadLn(fp, Buffer);
    Log.LogStatus('Reading NumFrames', 'uSprite.Init');
    sscanf(Buffer, 'FILES: %d', [@fNumframes]);
    Log.LogStatus('Reading NumFrames done', 'uSprite.Init');
    SetLength(fAnim, fNumframes);

    fBuilt := 1;
    Count  := 0;
    while(not EOF(fp)) and (Count < fNumframes) do
    begin
      ReadLn(fp, Buffer);
      if (Length(Buffer) = 0) then Continue;
      if (Copy(Buffer, 1,1 ) = '#') then Continue;
      Log.LogStatus('Reading name, pause, r, g, b', 'uSprite.Init');
      sscanf(Buffer, '%s %d %d %d %d', [@name, @pause, @r, @g, @b]);
      Log.LogStatus('Reading name, pause, r, g, b done', 'uSprite.Init');
      StrPCopy(Filename, aDir + '/' + Name);

      fAnim[Count] := TSpriteFrame.Create(Filename, Pause, r, g, b);
      if(fW <> 0) then begin
        fW := fAnim[Count].Image.w;
        if(fH <> 0) then begin
          fH := fAnim[Count].Image.W;
        end;
      end;

      Inc(Count);
    end;
  finally
    CloseFile(fp);
  end;
end;

// TSprite ---------------------------------------------------------------
procedure TSprite.SetAmimState(aState : TAnimState);
begin
  if (fLocked > 0) then Exit;

  Inc(fLocked);
  try
    if (fAnimState <> aState) then begin
      if Assigned(fAnimStateChange) then begin
        fAnimStateChange(Self, aState);
      end;
      fAnimState := aState;
    end;
  finally
    Dec(fLocked);
  end;
end;

function TSprite.GetWidth : Integer;
begin
  Result := fSpriteBase.Anim[0].Image.W;
end;

function TSprite.GetHeight : Integer;
begin
  Result := fSpriteBase.Anim[0].Image.H;
end;

function TSprite.GetRect : Types.TRect;
begin
  Result.Left   := Trunc(fX);
  Result.Top    := Trunc(fY);
  Result.Right  := Trunc(fX) + Width;
  Result.Bottom := Trunc(fY) + Height;
end;

function TSprite.GetSDLRect : TSDL_Rect;
begin
  Result.x := Trunc(fX);
  Result.y := Trunc(fY);
  Result.w := Width;
  Result.h := Height;
end;

function TSprite.GetCenterPos : TPoint;
begin
  Result.X := Trunc(fX + (Width div 2));
  Result.Y := Trunc(fY + (Height div 2));
end;

constructor TSprite.Create;
begin
  inherited Create;

  fLocked    := 0;
  fMoveSpeed := 1.0;
  fAnimSpeed := 1.0;
  fFrame     := 0;
  fLives     := 1;
  fVisible   := False;
  fAnimState := asNone;
end;

destructor TSprite.Destroy;
begin
  inherited Destroy;
end;

function TSprite.Init(aBase : TSpriteBase; aScreen : PSDL_Surface) : Integer;
begin
  fSpriteBase := aBase;
  if (fSpriteBase.fBuilt = 1) then begin
    if (fSpriteBase.fNumframes > 1) then begin
      fAnimating := True;
    end;
    fBackreplacement := SDL_DisplayFormat(fSpriteBase.Anim[0].Image);
  end;
  fScreen := aScreen;
  Result  := 0;
end;

procedure TSprite.Draw;
var
  Dest             : TSDL_Rect;
begin
  if fAnimating and (fAnimSpeed > 0.001) then begin
    if ((fLastupdate + fSpriteBase.Anim[fFrame].Pause * (1 / fAnimSpeed)) < SDL_GetTicks())
    then begin
      Inc(fFrame);
      if (fFrame > Pred(fSpriteBase.NumFrames)) then begin
        SetAmimState(asAnimStop);
        fFrame := 0;
      end;
      fLastupdate := SDL_GetTicks();
      SetAmimState(asAnimNext);
    end;
  end;

  if Visible then begin
    if (fDrawn = 0) then begin
      fDrawn := 1;
    end;

    Dest.X := Trunc(fX);
    Dest.Y := Trunc(fY);
    SDL_BlitSurface(fSpriteBase.Anim[fFrame].Image, nil, fScreen, @Dest);
  end;
end;

procedure TSprite.clearBG;
var
  Dest             : TSDL_Rect;
begin
  if (fDrawn = 1) then begin
    Dest.X := Trunc(fOldX);
    Dest.Y := Trunc(fOldY);
    Dest.W := fSpriteBase.Width;
    Dest.H := fSpriteBase.Height;
    SDL_BlitSurface(fBackreplacement, NIL, fScreen, @dest);
  end;
end;

procedure TSprite.updateBG;
var
  SrcRect          : TSDL_Rect;
begin
  SrcRect.x := Trunc(fX);
  SrcRect.y := Trunc(fY);
  SrcRect.w := fSpriteBase.Width;
  SrcRect.h := fSpriteBase.Height;
  fOldX     := fX;
  fOldY     := fY;
  SDL_BlitSurface(fScreen, @SrcRect, fBackreplacement, NIL);
end;

procedure TSprite.ToggleAnim;
begin
  fAnimating := not fAnimating;
end;

procedure TSprite.StartAnim;
begin
  fAnimating  := True;
  fLastupdate := SDL_GetTicks();
  SetAmimState(asAnimStart);
end;

procedure TSprite.StopAnim;
begin
  fAnimating := False;
  SetAmimState(asAnimStop);
end;

procedure TSprite.Rewind;
begin
  fFrame      := 0;
  fLastupdate := SDL_GetTicks();
end;

function TSprite.SpriteHit(aSprite : TSprite) : Boolean;
var
  SR1, SR2         : Types.TRect;
  outRect          : Types.TRect;
begin
  Result := False;
  if (aSprite <> Self) then begin
    SR1    := Self.Rect;
    SR2    := aSprite.Rect;
    if not IntersectRect(outRect, SR1, SR2) then begin
      // Keine Ueberschneidung, somit keine weitere Betrachtung notwendig
      Exit;
    end;

    Result := True;
  end;
end;

function TSprite.SpriteHitPixel(aSprite : TSprite) : Boolean;



var
  rightSprite      : TSprite;
  leftSprite       : TSprite;
  intersectWidth   : Integer;
  intersectHeight  : Integer;
  LSPoint          : TPoint;
  RSPoint          : TPoint;
  LSWidth          : Integer;
  LSHeight         : Integer;
  RSWidth          : Integer;
  RSHeight         : Integer;
  LSColorKey       : UINT32;
  RSColorKey       : UINT32;
  iX, iY           : Integer;
  LSSurface        : PSDL_Surface;
  RSSurface        : PSDL_Surface;

label DoNext;
begin 	
  Result := False;
  if not SpriteHit(aSprite) then begin
    Exit;
  end;

	// Determine which sprite is to the left and which is to the right.
	if (aSprite.X > X ) then begin
		rightSprite := aSprite;
		leftSprite  := Self;
	end else begin
		rightSprite := Self;
		leftSprite  := aSprite;
	end;

	// Get the width and height of each sprite
	LSWidth  := leftSprite.Width;
	LSHeight := leftSprite.Height;

	RSWidth  := rightSprite.Width;
	RSHeight := rightSprite.Height;

	// Get the amount the two sprites overlap in the X direction
	intersectWidth := LSWidth - (Trunc(rightSprite.X) - Trunc(leftSprite.X));

	// The intersecting width cannot be greater then the sprite width.
	if (intersectWidth > RSWidth) then begin
		intersectWidth := RSWidth;
  end;

	// Determine which sprite is lower on the screen.
	if (leftSprite.Y > rightSprite.Y) then begin
		LSPoint.Y := 0;
		RSPoint.Y := Trunc(leftSprite.Y) - Trunc(rightSprite.Y);
		intersectHeight := RSHeight - (Trunc(leftSprite.Y) - Trunc(rightSprite.Y));
		if (intersectHeight > LSHeight) then begin
			intersectHeight := LSHeight;
    end;
	end else begin
    // left sprite is higher on the screen. 
		LSPoint.Y := Trunc(rightSprite.Y) - Trunc(leftSprite.Y);
		RSPoint.Y := 0;
		intersectHeight := LSHeight - (Trunc(rightSprite.Y) - Trunc(leftSprite.Y));
		if (intersectHeight > RSHeight) then begin
			intersectHeight := RSHeight;
    end;
	end; 

	// Calculate the starting X position of collision.
	LSPoint.x := Trunc(rightSprite.X) - Trunc(leftSprite.X);
	RSPoint.x := 0;

  // Getting Surfaces
  LSSurface := leftSprite.SpriteBase.Anim[leftSprite.Frame].Image;
  RSSurface := rightSprite.SpriteBase.Anim[rightSprite.Frame].Image;

	// Get the color key for each surface.
	LSColorKey := LSSurface^.Format^.ColorKey;
	RSColorKey := RSSurface^.Format^.ColorKey;

  if (SDL_LockSurface(LSSurface) <> 0) then Exit;
  try
    if (SDL_LockSurface(RSSurface) <> 0) then Exit;
    try
      iY := 0;
      while (iY < intersectHeight) do begin
        iX := 0;
        while (iX < intersectWidth) do begin
          if (SDL_GetPixel(LSSurface, LSPoint.X + Trunc(iX),  LSPoint.Y + Trunc(iY)) = LSColorKey) then Goto DoNext;
          if (SDL_GetPixel(RSSurface, RSPoint.X + Trunc(iY),  RSPoint.Y + Trunc(iY)) = RSColorKey) then Goto DoNext;

          Result := True;
          Exit;
DoNext:
          Inc(iX);
        end;

        Inc(iY);
      end;
    finally
      SDL_UnlockSurface(RSSurface);
    end;
  finally
    SDL_UnlockSurface(LSSurface);
  end;
end;

procedure TSprite.SetCenter(X, Y : Double);
begin
  fX := X - (Width div 2);
  fY := Y - (Height div 2);
end;

procedure TSprite.SetCenter(Pos : TPoint);
begin
  fX := Pos.X - (Width div 2);
  fY := Pos.Y - (Height div 2);
end;

procedure TSprite.SetXY(X, Y : Double);
begin
  fX := X;
  fY := Y;
end;

procedure TSprite.XAdd(dX : Double);
begin
  fX := fX + dX;
end;

procedure TSprite.YAdd(dY : Double);
begin
  fY := fY + dY;
end;

end.

